/* ----------------------------------------------------------

Simple scan program for GRETINA engineering runs,
 data stream type 2  (after signal decomposition).

D.C. Radford 2011-4-6

Reads one or more input event data files
Outputs multiple spectra (.spe) and matrices (.spn)

-----------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <netinet/in.h>
#include <string.h>
#include <signal.h>

/*----------------------------------*/
/* spectra and matrices */

#define SPEC_SIZE 16384

int tot_proj[SPEC_SIZE];
int cc_proj[SPEC_SIZE];

int cc_id_spec[4096];
int time_spec[4096];
int deltatime_spec[4096];
int log_dt_spec[4096];
int int_mult_spec[4096];
int cc_mult_spec[4096];

int cc_C2_mat[100][4096];
int cc_E_mat[100][4096];
int cc_T_mat[500][4096];
int gg_mat[4096][4096];
int cc_id_mat[100][4096]; 
int int_mult_mat[120][4096]; 
int debug[100][4096];

void writeSpectra(const char *filename, int *buf, int dim);

/* ======================================================================= */

/* used for trapping control-C events, to interrupt scan */
int  gotsignal;
void breakhandler(int dummy) {
  gotsignal = 1;
}

/* ------------------------------------ */

int main(int argc, char **argv) {
  
  FILE *f, *fd, *matfile;  char  s[256], *pt;
  int   numEvts, numBufs, xtal, delta_time;
  int   cc_mult = 0, int_mult = 0, dt_old = 0;
  int   totE, ldt, lowE=0, cc_count=0;
  int   event[1000][2], cc_event[100][3];
  int   i, j, k;
  long long unsigned int time, time_old = 0, time_start = -1;

#define MAX_INTPTS 16
  struct crys_intpts {
    int   num;
    int   crystal_id;
    float tot_e;
    float t0;
    float chisq;
    float norm_chisq;
    long long int timestamp;
    struct {
      float x, y, z;
      float e;
    } intpts[MAX_INTPTS];
  } buf;

  static int t0x[20]={
    101,  91, 81,  95, 40, 37, 41, 31,  7, 27,
     39,  17, 82,  72, 92, 95, 80, 96, 76, 97};


  printf("\n Simple GRETINA sort, for stage 2 data...\n");

  /* open input file */
  if (argc < 2) {
    fprintf(stderr, "Usage: %s <filename>\n", argv[0]);
    return 1;
  }
  if ((fd = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "Error, cannot open file %s\n", argv[1]);
    return 1;
  }  

  // zero spectra and matrices
  bzero(tot_proj, sizeof(tot_proj));
  bzero(cc_proj, sizeof(cc_proj));
  bzero(cc_id_spec, sizeof(cc_id_spec));
  bzero(time_spec, sizeof(time_spec));
  bzero(deltatime_spec, sizeof(deltatime_spec));
  bzero(log_dt_spec, sizeof(log_dt_spec));
  bzero(int_mult_spec, sizeof(int_mult_spec));
  bzero(cc_mult_spec, sizeof(cc_mult_spec));

  bzero(cc_C2_mat[0], sizeof(cc_C2_mat));
  bzero(cc_E_mat[0], sizeof(cc_E_mat));
  bzero(cc_T_mat[0], sizeof(cc_E_mat));
  bzero(gg_mat[0], sizeof(gg_mat));
  bzero(cc_id_mat[0], sizeof(cc_id_mat));
  bzero(int_mult_mat[0], sizeof(int_mult_mat));
  bzero(debug[0], sizeof(debug));

  /* -------- read the name of the data file from input file -------- */

  // start of loop reading data file names
  while (fgets(s, 256, fd) != NULL
	 && s[0] != ' ' && s[0] != '\n' && s[0] != '\0' && s[0] != '\t') { 

    printf("\n Reading from data file %s\n"
	   " Press control-C to interrupt scan.\n", s);
    gotsignal = 0;
    signal(SIGINT, breakhandler);
 
    if ((pt = strchr(s , '\n'))) *pt = 0;  // remove trailing \n
    if (!(f = fopen(s, "r"))) {
      fprintf(stderr, "Error, cannot open file %s\n", s);
      return 1;
    }      
    numBufs = numEvts = 0; 

    /* -------- read a data buffer from the input file -------- */
    // start of loop reading data buffers
    while (!gotsignal &&
	   fread(&buf, sizeof(buf), 1, f) == 1)  {

      if ((++numBufs)%50000 == 0 || numBufs == 1) printf(" buff %d\n", numBufs);
      
      /* = = = = = = = = =   SORTING STARTS HERE   = = = = = = = = = */

      if (gotsignal) break;

      xtal = buf.crystal_id - 104;
      time = buf.timestamp;
      if (time_start == -1) time_start = time;  // first event
      totE = buf.tot_e;
      // printf("xtal, e: %3d %6d\n", xtal, totE);

      /* increment event-by-event spectra */
      if (totE < 0) {
	tot_proj[SPEC_SIZE-2]++;
      } else if (totE < SPEC_SIZE-2) {
	tot_proj[totE]++;
      } else {
	tot_proj[SPEC_SIZE-1]++;
      }

      delta_time = abs(time - time_old);
      ldt = 10.0 * log(delta_time + 1);
      if (totE > 20) {
	if (delta_time < 4096) {
	  deltatime_spec[delta_time]++;
	} else {
	  deltatime_spec[4095]++;
	}
	if (ldt < 4096) {
	  log_dt_spec[ldt]++;
	} else {
	  log_dt_spec[4095]++;
	}
      }

      /* increment histogram of signal-multiplicity per xtal */
      int_mult_mat[xtal][buf.num]++;

      if (delta_time > 100) {
	/* = = = =  We have a complete event  = = = = */

	/* increment histograms concerning multiplicity per event */
	if (int_mult > 1000 || cc_mult > 28) {
	  printf("Bad mult: %d %d\n", int_mult, cc_mult);
	  return 1;
	}
	int_mult_spec[int_mult]++;
	cc_mult_spec[cc_mult]++;
	if (cc_mult < 10) {
	  int_mult_mat[cc_mult + 100][int_mult]++;
	}

	/* Create matrices to show xtl coincidence events */
	// cc_event[x][0] = xtl number of event x
	// cc_event[x][1] = energy in keV 
	for (j = 0; j < cc_mult; j++) {
	  for (k = 0; k < cc_mult; k++) {
	    if (j == k) continue;
	    if (cc_event[j][1] > 400 && cc_event[k][1] > 400) {
	      ldt = 2000 + cc_event[j][2] - cc_event[k][2] -
		    t0x[cc_event[j][0]] + t0x[cc_event[k][0]];
	      cc_T_mat[cc_event[j][0]][ldt]++;
	      cc_T_mat[100+20*cc_event[j][0]+cc_event[k][0]][ldt]++;
	    }
	    if (cc_event[j][1] > 0 && cc_event[j][1] < 4096 &&
		cc_event[k][1] > 0 && cc_event[k][1] < 4096) {
	      gg_mat[cc_event[j][1]][cc_event[k][1]]++;
	      cc_id_mat[cc_event[j][0]][cc_event[k][0]]++;
	    }
	  }
	}

	if (delta_time > 1000 && delta_time < 2000) debug[0][cc_mult]++;
	if (delta_time > 3000 && delta_time < 4000) debug[2][cc_mult]++;
	if (dt_old > 1000 && dt_old < 2000) debug[1][cc_mult]++;
	if (dt_old > 3000 && dt_old < 4000) debug[3][cc_mult]++;

	dt_old = delta_time;

	/* = = = =  end of "complete event" processing = = = = */
	/*          initialize for next event                  */
	for (j = 0; j < 20; j++) cc_event[j][0] = cc_event[j][1] = -1;
	int_mult = cc_mult = 0;
	time_start = time;  // first time in new event
	numEvts++;
      }

      if (totE > 20) {
	int_mult += buf.num;
	for (i = 0; i < buf.num; i++){
	  event[int_mult][0] = xtal;
	  event[int_mult][1] = buf.intpts[i].e;
	  if (buf.intpts[i].e < 0) {
	    tot_proj[SPEC_SIZE-2]++;
	  } else if (buf.intpts[i].e < SPEC_SIZE-2) {
	    tot_proj[(int) buf.intpts[i].e]++;
	  } else {
	    tot_proj[SPEC_SIZE-1]++;
	  }
	}
      }
           
      /* central contact stuff */
      if (xtal < 0 || xtal >= 100) {
	printf("Bad xtal: %d\n", xtal);
	return 1;
      }
      cc_event[cc_mult][0] = xtal;  // xtal ID
      cc_event[cc_mult][1] = totE;  // xtal E
      // cc_event[cc_mult][2] = 10 * (time - time_start);
      cc_event[cc_mult][2] = 10 * (time - time_start) +
	                     (int) (10.0 * buf.t0); // time in ns
      //                     (int) (10.0 * buf.intpts[15].z); // time in ns
      cc_mult++;
      cc_count++;
      if (totE < 20) lowE++;
      cc_id_spec[xtal]++;
      if (totE < 0) {
	cc_proj[SPEC_SIZE-2]++;
      } else if (totE < SPEC_SIZE-2) {
	cc_proj[totE]++;
      } else {
	cc_proj[SPEC_SIZE-1]++;
      }
      if (totE < 4096 && totE > 0) {
	cc_E_mat[xtal][totE]++;
      }
      if (buf.chisq < 100.0) {
	cc_C2_mat[xtal][(int) (buf.chisq * 10.0f)]++;
      }
      if (buf.norm_chisq < 100.0) {
	cc_C2_mat[xtal][1000 + (int) (buf.norm_chisq * 10.0f)]++;
      }
      cc_C2_mat[xtal][2000 + (int) (buf.t0 * 100.0f)]++;

      delta_time = time - time_start + 100;
      if (totE > 20) {
	if (delta_time < 4096) {
	  time_spec[delta_time]++;
	} else {
	  time_spec[4095]++;
	}
      }

      if (buf.num < 15) {
	debug[10][100 + (int) (10.0 * buf.intpts[15].z)]++;
	debug[11][100 + (int) (10.0 * (buf.t0 - buf.intpts[15].z))]++;
	debug[12][100 + (int) (10.0 * buf.t0)]++;
	if (totE > 500) {
	  debug[20][100 + (int) (10.0 * buf.intpts[15].z)]++;
	  debug[21][100 + (int) (10.0 * (buf.t0 - buf.intpts[15].z))]++;
	  debug[22][100 + (int) (10.0 * buf.t0)]++;
	}
      }

      time_old = time;
    }  // end of loop reading data buffers

    /* = = = =  SORTING of one file ENDS HERE  = = = = */

    signal(SIGINT, SIG_DFL);    
    printf("\n %d buffers, %d events, %d low-E evts, %d CC-evts\n",
	   numBufs, numEvts, lowE, cc_count);
    
    /* write spectra */
    writeSpectra("cc_id.spe", cc_id_spec, 100);
    writeSpectra("time.spe", time_spec, 4096);
    writeSpectra("deltatime.spe", deltatime_spec, 4096);
    writeSpectra("log_dt.spe", log_dt_spec, 4096);
    writeSpectra("int_mult.spe", int_mult_spec, 4096);
    writeSpectra("cc_mult.spe", cc_mult_spec, 4096);
    writeSpectra("tot_proj.spe", tot_proj, SPEC_SIZE);
    writeSpectra("cc_proj.spe", cc_proj, SPEC_SIZE);
 
    /* write matrices */
    matfile = fopen("cc_c2_mat.spn", "wb");
    fwrite(cc_C2_mat[0], 1, sizeof(cc_C2_mat), matfile);
    fclose(matfile);

    matfile = fopen("cc_e_mat.spn", "wb");
    fwrite(cc_E_mat[0], 1, sizeof(cc_E_mat), matfile);
    fclose(matfile);

    matfile = fopen("cc_t_mat.spn", "wb");
    fwrite(cc_T_mat[0], 1, sizeof(cc_T_mat), matfile);
    fclose(matfile);

    matfile = fopen("gg_mat.spn", "wb");
    fwrite(gg_mat[0], 1, sizeof(gg_mat), matfile);
    fclose(matfile);

    matfile = fopen("cc_id_mat.spn", "wb");
    fwrite(cc_id_mat[0], 1, sizeof(cc_id_mat), matfile);
    fclose(matfile);

    matfile = fopen("int_mult_mat.spn", "wb");
    fwrite(int_mult_mat[0], 1, sizeof(int_mult_mat), matfile);
    fclose(matfile);

    matfile = fopen("debug.spn", "wb");
    fwrite(debug[0], 1, sizeof(debug), matfile);
    fclose(matfile);

    fclose(f); /* close the data file */

  }  // end of loop reading data file names (while(fgets))
  fclose(fd); /* close the main input file */

  return 0;
} // end main

/* ------------------------------------------ */

struct spePrefix {
  int reclA;
  char title[8];
  int dim;
  int a1;
  int a2;
  int a3;
  int reclB;
} spePre = {24, "", 0, 1, 1, 1, 24};

void writeSpectra(const char *filename, int *buf, int dim) {

  FILE  *fp_spectra;
  int   i, recl;
  float *fbuf;

  fp_spectra = fopen(filename, "w");
  if (fp_spectra == NULL) {
    fprintf(stderr, "unable to open file for spectra");
    exit(1);
  }

  strncpy(spePre.title, filename, 8);
  spePre.dim = dim;
  recl = sizeof(float) * dim;
  fbuf = (float *) calloc(dim, sizeof(float));
  if (fbuf == NULL) {
    fprintf(stderr, "calloc error\n");
    exit(2);
  }
  for (i = 0; i < dim; i++) {
    fbuf[i] = (float) buf[i];
  }

  (void) fwrite(&spePre, sizeof(struct spePrefix), 1, fp_spectra);
  (void) fwrite(&recl, sizeof(int), 1, fp_spectra);
  (void) fwrite(fbuf, sizeof(float), dim, fp_spectra);
  (void) fwrite(&recl, sizeof(int), 1, fp_spectra);

  fclose(fp_spectra);
  free(fbuf);
} /* writeSpectra() */
