/* ----------------------------------------------------------

Simple scan program for GRETINA engineering runs,
 data stream type 3  (before signal decomposition).

D.C. Radford 2011-4-6
Based on codes from P. Fallon -> C. Beasuang -> Aaron -> DCR

Reads one or more input event data files
Outputs multiple spectra (.spe) and matrices (.spn)

  -----------------------------------------------------------

To direct the program to the data, create an ascii file
   with the address of the the Global.dat file.
   You could do this for example by
$ echo /global/data2/gretina/Cave4CDec10/Run002/Global.dat >> in.dat

Then run the compiled code with the filename as a command line argument, eg
./scan_3_dcr in.dat

par_num is a way of identifying segments and central contacts,
  and is calculated using the formula
par_num = 500*(hole - 26) + 100*xtl + 10*boardId + chanId;

chanId 9 is missing from all boards 0 and 1,
  and refers to central contacts in the rest.

Boards take values 0-3, to make 4 per xtl

Hole equates to modules as follows,
something needs to be done when Q6 and Q7 are installed
hole_num   Q
  26       1
  27       4
  28       5
  29       2       
  30       3
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <netinet/in.h>
#include <string.h>
#include <signal.h>

// #include "sort.h"

#define MAXBUF 50000000            // number of input data buffers to process
#define DEG_TO_RAD 0.017453293
#define MAX_TraceLen 186
#define TraceLen 186
#define BUF_SIZE 40400 // 812900   // buffer size in bytes

#define SWAP 1

/*----------------------------------*/
/* spectra and matrices */

#define SPEC_SIZE 16384

int tot_proj[SPEC_SIZE];
int cc_proj[SPEC_SIZE];

int hole_spec[4096];
int crystal_spec[4096];
int board_spec[4096];
int channel_spec[4096];
int par_spec[2][4096];      // without & with energy threshold
int cc_id_spec[2][4096];    // without & with energy threshold
int time_spec[4096];
int deltatime_spec[4096];
int log_dt_spec[4096];
int sig_mult_spec[2][4096]; // without & with energy threshold
int cc_mult_spec[2][4096];  // without & with energy threshold
int debug_spec[4096];

int cc_E_mat[100][4096];
int par_E_mat[4096][4096];
int gg_mat[4096][4096];
int cc_id_mat[100][4096]; 
int par_id_mat[4096][4096]; 
int sig_mult_mat[120][4096]; 

/* energy gains; format: hole, crystal, board, channel, offset, gain */
float ge_gains[4096][6];

#define RANDNUM ((rand() & 4095) / 4096.0 - 0.5)
/* returns a random number between -0.5 and 0.5 */

void writeSpectra(const char *filename, int *buf, int dim);

/* ======================================================================= */

/* used for trapping control-C events, to interrupt scan */
int  gotsignal;
void breakhandler(int dummy) {
  gotsignal = 1;
}

/* ------------------------------------ */

int main(int argc, char **argv) {
  
  FILE *f, *fd, *f_gains, *matfile;
  char  s[256], *pt;
  int   numEvts, numBufs, sign;
  int   fullE = 0, chanId, holeNum, xtl, boardId;
  int   par_num, hole, cc_num, delta_time, linecount = 0, chn_no;
  int   sig_mult = 0, cc_mult = 0, sig_mult_wE = 0, cc_mult_wE = 0;
  int   sig_mult_xtal = 0, sig_mult_xtal_wE = 0, xtal_old = -1;
  int   compE, ldt, lowE=0, cc_count=0;
  int   event[1000][2], cc_event[100][2], pnsave[1000];
  int   i, j, k, TraceLen_read;
  float offset, gain;
  long long unsigned int time, time_old = 0, time_start = -1;
  unsigned short int     buf0[BUF_SIZE], buf1[BUF_SIZE], *evt;


  printf("\n Simple GRETINA sort, for stage 3 data...\n");

  /* open input file */
  if (argc < 2) {
    fprintf(stderr, "Usage: %s <filename>\n", argv[0]);
    return 1;
  }
  if ((fd = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "Error, cannot open file %s\n", argv[1]);
    return 1;
  }  

  fflush(stdout); // zero spectra and matrices
  bzero(tot_proj, sizeof(tot_proj));
  bzero(cc_proj, sizeof(cc_proj));
  bzero(hole_spec, sizeof(hole_spec));
  bzero(crystal_spec, sizeof(crystal_spec));
  bzero(board_spec, sizeof(board_spec));
  bzero(channel_spec, sizeof(channel_spec));
  bzero(par_spec[0], sizeof(par_spec));
  bzero(cc_id_spec[0], sizeof(cc_id_spec));
  bzero(time_spec, sizeof(time_spec));
  bzero(deltatime_spec, sizeof(deltatime_spec));
  bzero(log_dt_spec, sizeof(log_dt_spec));
  bzero(sig_mult_spec[0], sizeof(sig_mult_spec));
  bzero(cc_mult_spec[0], sizeof(cc_mult_spec));
  bzero(debug_spec, sizeof(debug_spec));

  bzero(cc_E_mat[0], sizeof(cc_E_mat));
  bzero(par_E_mat[0], sizeof(par_E_mat));
  bzero(gg_mat[0], sizeof(gg_mat));
  bzero(cc_id_mat[0], sizeof(cc_id_mat));
  bzero(par_id_mat[0], sizeof(par_id_mat));
  bzero(sig_mult_mat[0], sizeof(sig_mult_mat));

  /* Initialize all gains to 0 and 1 */
  for (hole = 26; hole < 31; hole++) {
    for (xtl=0; xtl < 4; xtl++) {
      for (boardId = 0; boardId < 4; boardId++) {
	for (chanId = 0; chanId < 10; chanId++) {
	  par_num = 500*(hole - 26) + 100*xtl + 10*boardId + chanId;
          ge_gains[par_num][0] = hole;
	  ge_gains[par_num][1] = xtl;
	  ge_gains[par_num][2] = boardId;
	  ge_gains[par_num][3] = chanId;
	  ge_gains[par_num][4] = 0.0;
	  ge_gains[par_num][5] = 1.0;
	  if((boardId==0 || boardId==1) && chanId==9) {
            ge_gains[par_num][4] = 0.0;
            ge_gains[par_num][5] = 0.0;
	  }
	}
      }
    }
  }

  /*
    read in the ge_gains from file name gains.txt
    in the following format:
    parnum gain offset\n
    an integer followed by two floats, all seperated only by whitespace
  */
  if (!(f_gains = fopen("gains.txt", "r"))) {
    fprintf(stderr, "Error, cannot open file gains.txt\n");
    return 1;
  }
  while (fscanf(f_gains, "%i%f%f\n", &par_num, &offset, &gain) == 3 &&
	 par_num >= 0 && par_num < 4096) {
    //  printf("%4d %7.4f %5.2\n", par_num, offset, gain);
    ge_gains[par_num][4] = gain;
    ge_gains[par_num][5] = offset;
    linecount++;
  }
  fclose(f_gains);
  printf ("Finished reading %d gains & offsets from gains.txt\n", linecount);

  /* -------- read the name of the data file from input file -------- */

  // start of loop reading data file names
  while (fgets(s, 256, fd) != NULL
	 && s[0] != ' ' && s[0] != '\n' && s[0] != '\0' && s[0] != '\t') { 

    printf("\n Reading from data file %s\n"
	   " Press control-C to interrupt scan.\n", s);
    gotsignal = 0;
    signal(SIGINT, breakhandler);
 
    if ((pt = strchr(s , '\n'))) *pt = 0;  // remove trailing \n
    if (!(f = fopen(s, "r"))) {
      fprintf(stderr, "Error, cannot open file %s\n", s);
      return 1;
    }      
    numBufs = numEvts = 0; 

    /* -------- read a data buffer from the input file -------- */
    // start of loop reading data buffers
    while (numBufs++ < MAXBUF && !gotsignal &&
	   fread(buf0, sizeof(short int), BUF_SIZE, f) == BUF_SIZE)  {

      if (numBufs%5000 == 0 || numBufs == 1) printf(" buff %d\n", numBufs);
      
      /* auto byte swap function*/
      /* do not use with file "central.dat"*/
      if (SWAP==1) {
	for (i = 0; i < BUF_SIZE; i++) {
	  buf0[i] = ntohs(buf0[i]);
	}
      }
      /* always swap order of shorts */
      for (i = 0; i < BUF_SIZE; i+=2) {
	buf1[i]   = buf0[i+1];
	buf1[i+1] = buf0[i];
      }
      
      /* = = = = = = = = =   SORTING STARTS HERE   = = = = = = = = = */

      // start of loop through buffer words
      for (i = 0; i < BUF_SIZE; i++) {
	/* make sure we start at an event header
	   and do not read past the end of a buffer */
	if (buf1[i] == 0xaaaa && buf1[i+1] == 0xaaaa &&
	    i+TraceLen+15 < BUF_SIZE) {  
	  numEvts++;
	  if (gotsignal) break;

	  evt = buf1 + i + 2;
	  TraceLen_read = 2*(evt[1] & 0x7FF) - 14;
     	  chanId  = evt[0] & 0xF;
	  holeNum = (evt[0] >> 8) & 0x1F;
	  xtl     = (evt[0] >> 6) & 0x3;
	  boardId = (evt[0] >> 4) & 0x3;
	  time = ((long long unsigned int) evt[2]) +
	        (((long long unsigned int) evt[3]) << 16) +
	        (((long long unsigned int) evt[4]) << 32);
	  if (time_start == -1) time_start = time;  // first event
	  chn_no = chanId + boardId*10;
	  hole = holeNum - 26; // CWB count holes from 0
	  par_num = hole*500 + xtl*100 + boardId*10 + chanId;
	  if (par_num < 0 || par_num >= 4000) {
	    printf("Bad par_num: %d\n", par_num);
	    return 1;
	  }

	  fullE = (((unsigned int) evt[6] & 0xFF)<<16) + (unsigned int) evt[5];
	  sign = (evt[6] & 0x0100);
	  if (sign) fullE = (int) (fullE - 0x01000000);
	  /* fix different polarity of cc and segments */
	  if (chn_no != 29 && chn_no != 39) {
	    fullE = -fullE;
	  }

	  compE = (ge_gains[par_num][4] +
		   ge_gains[par_num][5] * ((float) fullE) / 128.0
		   - 0.5);

	  /* increment event-by-event spectra */
	  hole_spec[holeNum]++;
	  crystal_spec[xtl]++;
	  board_spec[boardId]++;
	  channel_spec[chanId]++;
	  par_spec[0][par_num]++;
	  if (compE > 20) par_spec[1][par_num]++;
	  if (compE < 0) {
	    tot_proj[SPEC_SIZE-2]++;
	  } else if (compE < SPEC_SIZE-2) {
	    tot_proj[compE]++;
	  } else {
	    tot_proj[SPEC_SIZE-1]++;
	  }
	  if (compE < 4096 && compE >= 0) {
	    par_E_mat[par_num][compE]++;
	  }

	  delta_time = abs(time - time_old);
	  ldt = 10.0 * log(delta_time + 1);
	  if (compE > 20) {
	    if (delta_time < 4096) {
	      deltatime_spec[delta_time]++;
	    } else {
	      deltatime_spec[4095]++;
	    }
	    if (ldt < 4096) {
	      log_dt_spec[ldt]++;
	    } else {
	      log_dt_spec[4095]++;
	    }
	  }

	  if (xtal_old < 0) xtal_old = hole*4 + xtl;
	  if (xtal_old >= 0 &&
	      (xtal_old != hole*4 + xtl || delta_time > 100)) {
	    /* we have moved from one xtal event to another */
	    /* increment histograms concerning signal-multiplicity per xtal */
	    sig_mult_mat[xtal_old][sig_mult_xtal]++;
	    sig_mult_mat[xtal_old + 50][sig_mult_xtal_wE]++;
	    sig_mult_xtal = sig_mult_xtal_wE = 0;
	    xtal_old = hole*4 + xtl;
	  }

	  if (delta_time > 100) {
	    /* = = = =  We have a complete event  = = = = */

	    /* increment histograms concerning multiplicity per event */
	    if (sig_mult > 4000 || cc_mult > 28 || sig_mult_wE > 999) {
	      printf("Bad mult: %d %d %d\n",
		     sig_mult, cc_mult, sig_mult_wE);
	      return 1;
	    }
	    sig_mult_spec[0][sig_mult]++;
	    sig_mult_spec[1][sig_mult_wE]++;
	    cc_mult_spec[0][cc_mult]++;
	    cc_mult_spec[1][cc_mult_wE]++;
	    if (cc_mult < 10) {
	      sig_mult_mat[cc_mult + 100][sig_mult]++;
	      sig_mult_mat[cc_mult + 110][sig_mult_wE]++;
	    }

	    /* Create matrices to show xtl coincidence events */
	    // cc_event[x][0] = xtl number of event x
	    // cc_event[x][1] = energy in keV 
	    for (j = 1; j < cc_mult; j++) {
	      for (k = 0; k < j; k++) {
		if (cc_event[j][1] > 0 && cc_event[j][1] < 4096 &&
		    cc_event[k][1] > 0 && cc_event[k][1] < 4096) {
		  gg_mat[cc_event[j][1]][cc_event[k][1]]++;
		  gg_mat[cc_event[k][1]][cc_event[j][1]]++;
		  cc_id_mat[cc_event[j][0]][cc_event[k][0]]++;
		  cc_id_mat[cc_event[k][0]][cc_event[j][0]]++;
		}
	      }
	    }
	    for (j = 1; j < sig_mult_wE; j++) {
	      for (k = 0; k < j; k++) {
		par_id_mat[event[j][0]][event[k][0]]++;
		par_id_mat[event[k][0]][event[j][0]]++;
	      }
	    }

	    /* = = = =  end of "complete event" processing = = = = */
	    /*          initialize for next event                  */
	    for (j = 0; j < 20; j++) cc_event[j][0] = cc_event[j][1] = -1;
	    sig_mult = cc_mult = 0;
	    sig_mult_wE = cc_mult_wE = 0;
	    time_start = time;  // first time in new event
	  }

	  pnsave[sig_mult] = par_num;
	  sig_mult++;
	  sig_mult_xtal++;

	  if (compE > 20) {
	    event[sig_mult_wE][0] = par_num;
	    event[sig_mult_wE][1] = compE;
	    sig_mult_wE++;
	    sig_mult_xtal_wE++;
	  }
           
	  /* identify central contacts */
	  // if (boardId == 2 && chanId == 9 && cc_mult < 99) {
	  if (cc_mult < 99 &&
	      ((par_num%100 == 29 && par_num != 129 && par_num != 229) || 
	       par_num == 139 || par_num == 239)) {
	    cc_num = (int) hole*4 + xtl;
	    if (cc_num < 0 || cc_num >= 100) {
	      printf("Bad cc_num: %d\n", cc_num);
	      return 1;
	    }
	    cc_event[cc_mult][0] = cc_num;
	    cc_event[cc_mult][1] = compE;
	    cc_mult++;
	    if (compE > 20) cc_mult_wE++;
	    cc_count++;
	    if (compE < 20) lowE++;
	    cc_id_spec[0][cc_num]++;
	    if (compE > 20) cc_id_spec[1][cc_num]++;
	    if (compE < 0) {
	      cc_proj[SPEC_SIZE-2]++;
	    } else if (compE < SPEC_SIZE-2) {
	      cc_proj[compE]++;
	    } else {
	      cc_proj[SPEC_SIZE-1]++;
	    }
	    if (compE < 4096 && compE > 0) {
	      cc_E_mat[cc_num][compE]++;
	    }
	  }

	  delta_time = time - time_start + 100;
	  if (compE > 20) {
	    if (delta_time < 4096) {
	      time_spec[delta_time]++;
	    } else {
	      time_spec[4095]++;
	    }
	  }

	  time_old = time;
	  i += TraceLen_read + 15;
	}  // end of if (buf1[i] == 0xaaaa...
      }  // end of loop through buffer words
    }  // end of loop reading data buffers

    /* = = = =  SORTING of one file ENDS HERE  = = = = */

    signal(SIGINT, SIG_DFL);    
    printf("\n %d buffers, %d events, %d low-E evts, %d CC-evts\n",
	   numBufs, numEvts, lowE, cc_count);
    
    /* write spectra */
    writeSpectra("hole.spe", hole_spec, 100);
    writeSpectra("crystal.spe", crystal_spec, 100);
    writeSpectra("board.spe", board_spec, 100);
    writeSpectra("channel.spe", channel_spec, 100);
    writeSpectra("par.spe", par_spec[0], 4096);
    writeSpectra("par_wE.spe", par_spec[1], 4096);
    writeSpectra("cc_id.spe", cc_id_spec[0], 100);
    writeSpectra("cc_id_wE.spe", cc_id_spec[1], 100);
    writeSpectra("time.spe", time_spec, 4096);
    writeSpectra("deltatime.spe", deltatime_spec, 4096);
    writeSpectra("log_dt.spe", log_dt_spec, 4096);
    writeSpectra("sig_mult.spe", sig_mult_spec[0], 4096);
    writeSpectra("sig_mult_wE.spe", sig_mult_spec[1], 4096);
    writeSpectra("cc_mult.spe", cc_mult_spec[0], 4096);
    writeSpectra("cc_mult_wE.spe", cc_mult_spec[1], 4096);
    writeSpectra("tot_proj.spe", tot_proj, SPEC_SIZE);
    writeSpectra("cc_proj.spe", cc_proj, SPEC_SIZE);
    writeSpectra("debug.spe", debug_spec, 4096);
 
    /* write matrices */
    matfile = fopen("cc_e_mat.spn", "wb");
    fwrite(cc_E_mat[0], 1, sizeof(cc_E_mat), matfile);
    fclose(matfile);

    matfile = fopen("par_e_mat.spn", "wb");
    fwrite(par_E_mat[0], 1, sizeof(par_E_mat), matfile);
    fclose(matfile);

    matfile = fopen("gg_mat.spn", "wb");
    fwrite(gg_mat[0], 1, sizeof(gg_mat), matfile);
    fclose(matfile);

    matfile = fopen("cc_id_mat.spn", "wb");
    fwrite(cc_id_mat[0], 1, sizeof(cc_id_mat), matfile);
    fclose(matfile);

    matfile = fopen("par_id_mat.spn", "wb");
    fwrite(par_id_mat[0], 1, sizeof(par_id_mat), matfile);
    fclose(matfile);

    matfile = fopen("sig_mult_mat.spn", "wb");
    fwrite(sig_mult_mat[0], 1, sizeof(sig_mult_mat), matfile);
    fclose(matfile);

    fclose(f); /* close the data file */

  }  // end of loop reading data file names (while(fgets))
  fclose(fd); /* close the main input file */

  return 0;
} // end main

/* ------------------------------------------ */

struct spePrefix {
  int reclA;
  char title[8];
  int dim;
  int a1;
  int a2;
  int a3;
  int reclB;
} spePre = {24, "", 0, 1, 1, 1, 24};

void writeSpectra(const char *filename, int *buf, int dim) {

  FILE  *fp_spectra;
  int   i, recl;
  float *fbuf;

  fp_spectra = fopen(filename, "w");
  if (fp_spectra == NULL) {
    fprintf(stderr, "unable to open file for spectra");
    exit(1);
  }

  strncpy(spePre.title, filename, 8);
  spePre.dim = dim;
  recl = sizeof(float) * dim;
  fbuf = (float *) calloc(dim, sizeof(float));
  if (fbuf == NULL) {
    fprintf(stderr, "calloc error\n");
    exit(2);
  }
  for (i = 0; i < dim; i++) {
    fbuf[i] = (float) buf[i];
  }

  (void) fwrite(&spePre, sizeof(struct spePrefix), 1, fp_spectra);
  (void) fwrite(&recl, sizeof(int), 1, fp_spectra);
  (void) fwrite(fbuf, sizeof(float), dim, fp_spectra);
  (void) fwrite(&recl, sizeof(int), 1, fp_spectra);

  fclose(fp_spectra);
  free(fbuf);
} /* writeSpectra() */
